# -*- coding: utf-8 -*-
"""
Le manager s'occupe de lire la config et de lancer
les différents workers
"""

# Imports généraux
from __future__ import absolute_import
import sys
import os
import codecs
import traceback
# Pour le fichier de config
import six.moves.configparser
# Pour les logs
import logging
import logging.handlers
# Nos propres modules
from everything.workers.advia import WorkerAdvia
from everything.workers.dms import WorkerDms
from everything.workers.accueil import WorkerAccueil
from everything.workers.ctlink import WorkerCtlink
from everything.workers.sysmex import WorkerSysmex, WorkerSysmexSFTP
from everything.workers.ftp import WorkerFTP
from everything.workers.sftp import WorkerSFTP
from everything.workers.samba import WorkerSamba
from everything.workers.capi import WorkerCapi
from everything.workers.swing import WorkerSwing
from everything.workers.img import WorkerImgSplit
from everything.workers.autovue import WorkerAutovue
from everything.workers.garbage import GarbageCollector
from everything.constants import ENCODING
import six


class Manager(object):
    """
    La classe Manager (voir le docstring du module)
    """
    default_config = {
        "actif": '0',
        "mca_version": '4.x',
        "dbname": "auto",
        "user": "auto",
        "host": "localhost",
        "port": '5432',
        "passwd": "sol2",
        "code_test": "GCGR",
        "paillasse": "HEMATO",
        "num_automate": "0",
        "dirFich": '/home/ftp/advia2120',
        "minutes": '5',
        "age_max_demande": "1",
        "type": 'advia',
        "ftp_ok": "OK",
        "ftp_del": "1",
        "ftp_url": "ftp://user:password@hostname_or_ip/distant/folder/",
        "ftp_re": "",
        "ftp_timeout": "30",
        "ftp_recursive": "0",
        "search_subfolders": "",
        "sftp_url": "sftp://user:password@hostname_or_ip/distant/folder/",
        "sftp_re": "",
        "smb_ok": "OK",
        "smb_del": "1",
        "smb_url": "smb://user:password@hostname_or_ip/distant/folder/",
        "smb_domain": "WORKGROUP",
        "smb_use_flags": "1",
        "smb_re": "",
        "commentaire": "",
        "parse_pdf": "0",
        "fname_re": '',
        "transcodage": '',
        "use_dossier": "0",
        "compress_img": "0",
        "resize_img": "0",
        "outputdir": "/tmp/",
        "del_on_success": "0",
        "split_geometry": "",
        "age_fichier": "0",
        "create_ok": "0",
        "shadoc_url": "",
        "validation_testbio": "0",
        "type_fichier": "png",
        "ftp_actif": "0"
    }
    config_file_doc = u"""# ATTENTION: Version de MCA 3.2.14 ou superieure
#   requise.
# Configuration
# Créer une nouvelle section [...] pour chaque connexion à gérer.
# Le nom de la section est libre mais si vous ne dépassez pas 8 caractères,
#   vos logs seront plus jolis (mieux alignés).
# Ne recopier que les informations différentes des valeurs par défaut.
#
# Options :
#
# actif: mettre 1 pour activer cette section
# mca_version: [4.x] par défaut. Valeurs possibles :
#   3.x
#   4.x
#
# Connexion à la BDD :
#   dbname: nom de la BDD
#   user: login
#   host: IP ou nom d'hôte
#   port: port TCP
#   passwd: mot de passe
#
# Paramétrage:
#   code_test: code du test à utiliser pour les graphes
#   paillasse: code de la paillasse
#   num_automate: numéro de la connexion sur la paillasse
#   dirFich: répertoire du FTP où sont les fichiers
#   outputdir: répertoire de sortie
#   minutes: écart maximum entre la réception du fichier et du résultat
#   age_max_demande: âge maximum de la demande en jours. Il s'agit du nombre de
#       jours en plus du jour en cours (par défaut : 1 = on accepte des
#       fichiers sur des demandes depuis la veille au matin)
#   transcodage: permet de transcoder des codes tests lus dans le nom du
#       fichier
#       format: TESTLU^TESTMCA|TESTLU2^TESTMCA2
#   type: type de traitement à effectuer
#       voici les valeurs possibles :
#           advia  - advia 2120 en direct
#           ctlink - advia 2120 derrière une chaîne centralink
#           dms    - advia 2120 derrière une chaîne DMS
#           sysmex - sysmex
#           ftpget - récupération de fichiers sur un compte FTP distant
#           ftpput - dépôt de fichiers sur un compte FTP distant
#           sftpget - récupération de fichiers sur un compte SFTP distant
#           smbget - récupération de fichiers sur un compte SMB distant
#           smbput - déôt de fichiers sur un compte SMB distant
#           capi   - récupération de graphes Capillarys
#           swing  - récupération des PDFs du Swing
#           ruby   - Ruby
#           img_split - Découpe d'image
#           autovue - Récupération d'images Autovue
#   use_dossier: utiliser le numero du dossier au lieu du code barre quand on
#       cherche une correspondance
#   compress_img: compresser l'image
#   resize_img: redimensionner l'image (format: WIDTHxHEIGHT: 70x187). L'image
#       ne sera pas déformée et peut donc être au final plus petite que la
#       taille paramétrée
#   del_on_success: supprimer l'object après traitement
#   split_geometry: comment découper l'image (voir la DOC)
#   age_fichier: age minimum du fichier (en heures) avant suppression
#   create_ok: créer un fichier OK après avoir déposé un fichier local (mettre
#       le motif par lequel remplacer l'extension du fichier ou 0 pour
#       désactiver)
#
# FTP Get:
#   s'occupe de rapatrier des fichiers depuis un compte FTP distant vers le
#       répertoire local 'dirfich'.
#   Paramètres :
#       ftp_url: url d'accès au FTP
#       ftp_re: expression régulère pour la reconnaissance des noms de fichiers
#       ftp_ok: attend la présence d'un fichier OK (mettre le motif par lequel
#          remplacer l'extension du fichier ou 0 pour désactiver)
#       ftp_del: supprime le fichier après sa récupération
#       ftp_timeout: timeout en secondes (il y a de toute façon un timeout
#           général à 30)
#       ftp_recursive: doit-on recuperer les fichiers dans les sous dossiers
#       search_subfolders: Liste des dossiers dans lesquels chercher le fichier
#           (utile pour le worker sysmex)
#
# SFTP Get:
#   idem que FTP Get mais sur un SFTP
#
# Samba Get:
#   idem que FTP Get mais sur un partage windows
#       smb_domain: Groupe de travail Samba (WORKGROUP par défaut)
#       fname_re: expression régulière permettant de trouver le numéro de
#           dossier/code test dans le nom du fichier
# Samba Put:
#   idem que Samba Get mais avec une option permettant de gérer les conflits
#   entre les versions de libsmbclient et pycmbc :
#       smb_use_flags: utiliser les flags O_CREAT | O_WRONLY lors de
#           l'ouverture d'un fichier distant
#
# Commentaire:
#   il est possible de saisir un commentaire.
#
# Capillarys:
#   Paramètres :
#       parse_pdf: doit-on parser le PDF pour trouver le numéro de dossier
#           et du code test
#       fname_re: expression régulière permettant de trouver le numéro de
#           dossier/code test dans le nom du fichier
# Sysmex:
#   Paramètres:
#      validation_testbio: utile pour transmettre les fichiers intégrés au SIL
#      type_fichier: fixe le type de fichier intégré, png par defaut.
#
"""
    config_type_to_worker_class = {
        'advia': WorkerAdvia,
        'accueil': WorkerAccueil,
        'ctlink': WorkerCtlink,
        'sysmex': WorkerSysmex,
        'sysmexsftp': WorkerSysmexSFTP,
        'ftpget': WorkerFTP,
        'ftpput': WorkerFTP,
        'sftpget': WorkerSFTP,
        'smbget': WorkerSamba,
        'smbput': WorkerSamba,
        'capi': WorkerCapi,
        'swing': WorkerSwing,
        'img_split': WorkerImgSplit,
        'autovue': WorkerAutovue,
        'garbage': GarbageCollector,
        'dms': WorkerDms,
    }

    def __init__(self, conffile, verbosity=logging.DEBUG, logfile=None):
        self.conffile = conffile
        self.logfile = logfile
        self.workers = []
        self.logger = logging.getLogger(None)
        self.logger.setLevel(verbosity)
        if logfile is not None:
            rfh = logging.handlers.RotatingFileHandler(
                self.logfile,
                maxBytes=1 * 1024 * 1024,
                backupCount=5)
        else:
            rfh = logging.StreamHandler()
        rfh.setFormatter(logging.Formatter(
            "%(asctime)s - %(name)8s - %(levelname)5.5s - %(message)s"))
        self.logger.addHandler(rfh)

    def configure(self):
        """
        Fonction qui lit le fichier de config et instancie les différents
        workers
        """
        config = six.moves.configparser.RawConfigParser(self.default_config)
        if not os.path.exists(self.conffile):
            self.logger.fatal(u"Fichier de configuration absent")
            return False
        self.logger.debug(u"Lecture du fichier de configuration %s" , self.conffile)
        with codecs.open(self.conffile, encoding=ENCODING) as configfile:
            config.readfp(configfile)
        for section in config.sections():
            options = dict(config.items(section))
            if options.get("actif", '0') != '1':
                self.logger.warning(
                    u"%s - La section est désactivée. Section ignorée.",
                    section)
                continue

            if (
                options.get("type", '1') not in
                list(self.config_type_to_worker_class.keys())
            ):
                self.logger.warning(
                    u"%s - Le type %s spécifié dans la section n'existe pas. "
                    u"Section ignorée.",
                    section, options.get("type", 1))
                continue

            self.logger.debug(
                u"Initialisation et vérification de la configuration "
                u"du worker %s", section)
            worker = self.config_type_to_worker_class[options.get("type")](
                section, options)
            if worker.error:
                error_msg_suffix = u"Section ignorée."
                self.logger.warning(
                    u"%s - Erreur à l'initialisation pour la section : %s. %s",
                    section, worker.error, error_msg_suffix)
            else:
                self.workers.append(worker)
        return True

    def writeconf(self, cparser):
        """
        This is a unicode aware re-write of the config parser 'write' method
        See : /usr/lib/python2.6/ConfigParser.py
        """
        lines = self.config_file_doc.split(u"\n")
        if cparser._defaults:
            lines.append(u"[%s]" % six.moves.configparser.DEFAULTSECT)
            for (key, value) in cparser._defaults.items():
                lines.append(u"%s = %s" %
                             (key, six.text_type(value).replace('\n', '\n\t')))
            lines.append(u"")
        for section in cparser._sections:
            lines.append(u"[%s]" % section)
            for (key, value) in cparser._sections[section].items():
                if key != "__name__":
                    if isinstance(value, six.text_type):
                        value = value.replace('\n', '\n\t')
                    elif isinstance(value, str):
                        value = value.decode(ENCODING).replace('\n', '\n\t')
                    lines.append(u"%s = %s" % (key, value))
            lines.append(u"")
        with codecs.open(self.conffile,
                         encoding=ENCODING,
                         mode='wb') as configfile:
            configfile.write(u"\n".join(lines))

    def show(self, buff=sys.stdout, log=False):
        """
        Fonction utilisée pour afficher un worker (sa config)
        """
        output = []
        for worker in self.workers:
            output.append(u"#" * 80)
            output.append(worker.name.center(80))
            output.append(u"=" * 80)
            output.extend(worker.show(buff=None))
            output.append(u"#" * 80)
        output.append(u"")
        if buff:
            buff.write(u"\n".join(output).encode(ENCODING))
        if log:
            for line in output:
                self.logger.debug(line.encode(ENCODING))
        return output

    def run(self, filecount=None):
        """
        Lance les différents workers en leur demandant de traiter filecount
        fichiers
        """
        for worker in self.workers:
            try:
                worker.run(filecount)
            except:
                self.logger.error(traceback.format_exc())
            finally:
                worker.teardown()
