# -*- coding: utf-8 -*-
"""
L'import des graphes pour un Advia
"""

# Imports généraux
from __future__ import absolute_import
import os
import re
import datetime

from everything.workers.base import Worker
from everything.constants import MOVE_TO_ERROR, DONT_MOVE
from everything.exceptions import FileOrFilenamException


class WorkerAdvia(Worker):
    """
    cf module docstring
    """
    filename_regexp = re.compile('w(?P<connection_code>[0-9]{3})(?P<filenum>[0-9]{4})\.bmp')
    imgformat = 'png'
    subfolders = ['traite', 'erreur']
    need_bdd = True
    sql = """select
            da.automate,
            d.id,
            da.id,
            t.id,
            r.id,
            pr.id
        from
            demande d
            join dem_auto da on (d.id = da.demande)
            join dem_auto_test dat on (dat.dem_auto = da.id)
            join resultat r on (r.demande = d.id)
            join testbio t on (t.id = r.testbio and t.id = dat.testbio)
            join resu_auto ra on (ra.id = r.resu_auto and  ra.dem_auto = da.id)
            join prescription as pr on (pr.demande = d.id and pr.testbio = t.id)
        where
            da.automate = %%s and t.code= %%s
            and d.date_demande > now()::date - '%%s days'::interval
            and atofornull(r.resu) is not null
            and atofornull(r.resu)::integer = %%s
            and r.date_resu is not null
            and (age(r.date_resu, %%s) > '%%s minutes' and age(r.date_resu, %%s) < '%%s minutes')
            %(extra_where)s
        order by r.date_resu desc"""

    def get_extra_where(self):
        return ""

    def check_and_get_sql_params(self, filename):
        """
        Vérifie que le fichier est correct
        et retourne les paramètres à utiliser avec la requête SQL

        Surcharger cette méthode au besoin
        """
        match = self.filename_regexp.match(filename)
        if not match:
            raise FileOrFilenamException(
                u"%s - Le nom du fichier ne matche pas l'expression régulière %s." % (filename, self.filename_regexp.pattern))
        try:
            filenum = int(match.group('filenum'))
        except IndexError:
            raise FileOrFilenamException(
                u"%s - Le nom du fichier ne permet pas de trouver le numéro." % (filename,))

        self.logger.debug(u'%s - Valeur %s attendue (numéro du fichier) : %s',
                          filename, self.config['code_test'], filenum)
        datefichier = datetime.datetime.fromtimestamp(
            os.path.getmtime(os.path.join(self.config['dirfich'], filename)))
        self.logger.debug(u"%s - date du fichier : %s", filename, datefichier)
        return [
            self.config['paillasse'],
            self.config['code_test'],
            self.config['age_max_demande'],
            filenum,
            datefichier,
            -self.config['minutes'],
            datefichier,
            self.config['minutes'],
        ]

    def process(self, filename):
        try:
            params = self.check_and_get_sql_params(filename)
        except FileOrFilenamException as e:
            self.logger.error(e.msg)
            return MOVE_TO_ERROR
        datefichier = datetime.datetime.fromtimestamp(
            os.path.getmtime(os.path.join(self.config['dirfich'], filename)))
        extra_where = self.get_extra_where()
        if self.config['num_automate'] != 0:
            self.logger.debug(
                u"%s - Utilisation du numéro d'automate %s (%s)",
                filename, self.config['num_automate'], 2 ** self.config['num_automate'])
            extra_where += u" and ra.n_auto = %s "
            params.append(self.config['num_automate'])
        sql = self.sql % {'extra_where': extra_where, 'use_dossier': self.config['use_dossier']}
        self.logger.debug("REQUETE : %s %s", sql, params)
        cursor = self.conn.cursor()
        cursor.execute(sql, params)
        self.conn.commit()
        resultats_trouves = cursor.fetchall()
        if len(resultats_trouves) == 0:
            self.logger.error(
                u"%s - Impossible de trouver un résultat correspondant au fichier (vérifiez les paramètres et essayez d'augmenter le nombre de minutes d'écart tolérées)",
                filename)
            if datetime.datetime.now() - datefichier < datetime.timedelta(minutes=self.config['minutes']):
                self.logger.info(u"%s - Fichier assez jeune, attente", filename)
                return DONT_MOVE
            return MOVE_TO_ERROR
        elif len(resultats_trouves) > 1:
            self.logger.error(
                u"%s - Impossible de trouver un unique résultat correspondant au fichier (on en a trouvé %s)",
                filename, len(resultats_trouves))
            self.logger.debug(
                u"%s - Infos trouvées : %s",
                filename, resultats_trouves)
            self.logger.debug(u"%s - Paramètres de la requête : %s", filename, params)
            return MOVE_TO_ERROR

        resu = resultats_trouves[0]
        self.logger.debug(u"%s - Un résultat trouvé : %s", filename, resu)

        fichier = open(os.path.join(self.config['dirfich'], filename), 'rb')
        fichier.seek(0)
        bufimg = fichier.read()
        fichier.close()
        automate, demande, dem_auto, testbio, id_resultat, id_prescription = resu
        return self.save_img(
            filename, demande, dem_auto, testbio, bufimg,
            imgformat=self.imgformat, automate=automate,
            id_resultat=id_resultat, id_prescription=id_prescription)

    def list_files(self):
        """
        Liste les fichiers à traiter
        """
        flist = os.listdir(self.config['dirfich'])
        for special in self.subfolders:
            if special in flist:
                flist.remove(special)
        return flist
