# -*- coding: utf-8 -*-
"""
Le worker de base, tous les autres workers héritent de celui-ci
qui fournit des méthodes que tous utilisent.
"""

# Imports généraux
from __future__ import absolute_import
from __future__ import print_function
import re
import os
import io
import sys
import stat
import shutil
import tempfile
import paramiko
import psycopg2
import smbc
from PIL import Image
import logging
import logging.handlers
from ftplib import FTP, error_perm, all_errors
import socket
import time

try:
    import shadoc_client
except ImportError:
    print("Shadoc import failed")
    shadoc_client = None

# Nos propres imports
from everything.constants import (
    ENCODING,
    DONT_MOVE,
    MOVE_TO_ERROR,
    MOVE_TO_DONE,
    MOVE_TO_TRASH,
    MCA_VERSIONS_USERFRIENDLY,
)


class Worker(object):
    """
    Définition de la classe Worker
    """
    mapping_check = {
        'minutes': (
            lambda self, value:
            float(value), u"Le paramètre 'minutes' doit être un nombre"),
        'age_max_demande': (
            lambda self, value:
            float(value),
            u"Le paramètre 'age_max_demande' doit être un nombre"),
        'age_fichier': (
            lambda self, value:
            float(value), u"Le paramètre 'age_fichier' doit être un nombre"),
        'num_automate': (
            lambda self, value:
            int(value),
            u"Le paramètre 'num_automate' doit être un nombre entier"),
        'port': (
            lambda self, value:
            int(value), u"Le paramètre 'port' doit être un nombre entier"),
        'ftp_timeout': (
            lambda self, value:
            int(value),
            u"Le paramètre 'ftp_timeout' doit être un nombre entier"),
        'ftp_re': (
            lambda self, value:
            re.compile(value) if value else re.compile('.*'),
            u"Impossible de compiler la regexp 'ftp_re'"),
        'sftp_re': (
            lambda self, value:
            re.compile(value) if value else re.compile('.*'),
            u"Impossible de compiler la regexp 'sftp_re'"),
        'smb_re': (
            lambda self, value:
            re.compile(value) if value else re.compile('.*'),
            u"Impossible de compiler la regexp 'smb_re'"),
        'smb_use_flags': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'smb_use_flags' doit être 0 ou 1"),
        'fname_re': (
            lambda self, value:
            re.compile(value) if value else None,
            u"Impossible de compiler la regexp 'fname_re'"),
        'ftp_del': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'ftp_del' doit être 0 ou 1"),
        'smb_del': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'smb_del' doit être 0 ou 1"),
        'ftp_recursive': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'ftp_recursive' doit être 0 ou 1"),
        'ftp_actif': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'ftp_actif' doit être 0 ou 1"),
        'parse_pdf': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'parse_pdf' doit être 0 ou 1"),
        'transcodage': (
            lambda self, value:
            {} if not value else dict(
                [transcodage.split('^')
                 for transcodage in value.split('|')]),
            u"Le parametre 'transcodage' n'est pas correct"),
        'use_dossier': (
            lambda self, value:
            ["tube_auto", "tube"][int(value)],
            u"Le paramètre 'use_dossier' doit être 0 ou 1"),
        'compress_img': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'compress_img' doit être 0 ou 1"),
        'del_on_success': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'del_on_success' doit être 0 ou 1"),
        'split_geometry': (
            lambda self, value:
            [] if not value else [
                [int(v) for v in s.split(',')]
                for s in value.split('|')],
            u"Le paramètre 'split_geometry' n'est pas correct"),
        'mca_version': (
            lambda self, value: MCA_VERSIONS_USERFRIENDLY.index(value),
            u"La valeur du paramètre 'mca_version' est incorrecte"
        ),
        'resize_img': (
            lambda self, value:
            None if value == "0"
            else [int(V) for V in value.split("x")],
            u"La valeur du paramètre 'resize_img' est incorrecte"
        ),
        'shadoc_url': (
            lambda self, value: value,
            "",
        ),
        'validation_testbio': (
            lambda self, value:
            [False, True][int(value)],
            u"Le paramètre 'validation_testbio' doit être 0 ou 1"),
        'type_fichier': (
            lambda self, value: value,
            ""),
    }
    subfolders = []
    need_bdd = False
    need_ftp = False
    need_sftp = False
    need_samba = False

    def __init__(self, name, config):
        self.name = name
        self.config = config
        # Préparation de l'environnement de travail
        # La configuration
        self.error = self.check_config()
        if self.error:
            return
        # Logger
        self.logger = logging.getLogger(self.name)
        # Répertoires
        try:
            for subdir in self.subfolders:
                directory = os.path.join(self.config['dirfich'], subdir)
                if not os.path.exists(directory):
                    os.makedirs(directory)
            if self.config['outputdir']:
                if not os.path.exists(self.config['outputdir']):
                    os.makedirs(self.config['outputdir'])
        except:
            self.error = u"Impossible de créer les répertoires de travail"
            return
        # Connexion BDD
        self.conn = None
        if self.need_bdd:
            try:
                self.conn = psycopg2.connect(
                    database=self.config['dbname'],
                    user=self.config['user'],
                    password=self.config['passwd'],
                    host=self.config['host'],
                    port=self.config['port']
                )
            except:
                self.error = u"Connexion à la BDD impossible"
                return

    def check_config(self):
        """
        Vérifie la config du worker en appelant les différents lambda
        définies pour cela
        """
        for key, value in self.mapping_check.items():
            try:
                self.config[key] = value[0](self, self.config[key])
            except:
                return value[1]
        if self.need_ftp:
            try:
                return self.check_ftp_config()
            except:
                return u"Erreur non reconnue dans la config FTP"
        if self.need_sftp:
            try:
                return self.check_sftp_config()
            except:
                return u"Erreur non reconnue dans la config SFTP"
        if self.need_samba:
            try:
                return self.check_samba_config()
            except:
                return u"Erreur non reconnue dans la config SMB"

    def check_ftp_config(self):
        """
        Vérifie la configuration et la connexion au FTP
        """
        ftpurl_regexp = re.compile(
            'ftp://(?P<user>[^:]+):(?P<passwd>[^@]+)@(?P<hostorip>[^/]+)/' +
            '(?P<distant_folder>.*)')
        matches = ftpurl_regexp.match(self.config['ftp_url'])
        if not matches:
            return (u"Le paramètre 'ftp_url' ne correspond pas à l'expression "
                    u"régulière de vérification.")
        user, password, host, folder = matches.groups()
        try:
            ftp = FTP(host, user, password, "", self.config['ftp_timeout'])
        except socket.gaierror:
            return u"L'hôte %s est inconnu" % (host,)
        except socket.error:
            return u"Erreur de socket FTP (connexion refusée)"
        except error_perm:
            return u"Permission refusée sur le FTP"
        except socket.timeout:
            return u"Timeout sur la connexion FTP"
        except TypeError:
            return u"Mauvais paramétrage de connexion FTP"

        try:
            ftp.cwd(folder)
        except error_perm:
            return u"Le dossier %s n'existe pas sur ce FTP" % (folder,)
        self.config['ftp_object'] = ftp

    def check_sftp_config(self):
        """
        Vérifie la configuration et la connexion au SFTP
        """
        sftpurl_regexp = re.compile(
            'sftp://(?P<user>[^:]+):(?P<passwd>[^@]+)@(?P<hostorip>[^/]+)/' +
            '(?P<distant_folder>.*)')
        matches = sftpurl_regexp.match(self.config['sftp_url'])
        if not matches:
            return (u"Le paramètre 'sftp_url' ne correspond pas à "
                    u"l'expression régulière de vérification.")
        user, password, host, folder = matches.groups()
        try:
            sftp_transport = paramiko.Transport((host, 22))
            sftp_transport.connect(username=user, password=password)
            sftp = paramiko.SFTPClient.from_transport(sftp_transport)
        except socket.gaierror:
            return u"L'hôte %s est inconnu" % (host,)
        except socket.error:
            return u"Erreur de socket SFTP (connexion refusée)"
        except socket.timeout:
            return u"Timeout sur la connexion SFTP"
        except paramiko.SSHException as e:
            return u"SSHException: %s" % (e,)
        except TypeError:
            return u"Mauvais paramétrage de connexion SFTP"

        try:
            sftp.chdir(folder)
        except IOError:
            return u"Le dossier %s n'existe pas sur ce SFTP" % (folder,)
        self.config['sftp_object'] = sftp

    def check_samba_config(self):
        """
        Vérifie la configuration et la connexion Samba
        """
        smburl_regexp = re.compile(
            'smb://(?P<user>[^:]*):(?P<passwd>[^@]*)@(?P<hostorip>[^/]+)/' +
            '(?P<distant_folder>.*)')
        matches = smburl_regexp.match(self.config['smb_url'])
        if not matches:
            return (u"Le paramètre 'smb_url' ne correspond pas à "
                    u"l'expression régulière de vérification.")
        user, password, host, folder = matches.groups()

        def auth(*args):
            return (self.config['smb_domain'], user, password)
        ctx = smbc.Context(auth_fn=auth)
        try:
            ctx.opendir("smb://%s" % (host,)).getdents()
        except RuntimeError as e:
            return u"Erreur de connexion samba, %s" % (e.args,)
        except smbc.TimedOutError as e:
            return (u"Timeout lors de la connexion au serveur samba, %s" %
                    (e.args,))
        try:
            ctx.opendir("smb://%s/%s" % (host, folder,))
        except RuntimeError as e:
            return u"Impossible d'ouvrir le dossier distant, %s" % (e.args,)

        self.config['smb_full_url'] = "smb://%s/%s" % (host, folder,)
        self.config['smb_object'] = ctx

    def show(self, buff=sys.stdout, log=False):
        """
        Permet d'afficher les informations de ce worker à l'utilisateur
        """
        ignore = ['ftp_object', 'smb_object', ]
        output = []
        for option, value in self.config.items():
            if option not in ignore:
                output.append(u"%15.15s : %s" % (option, value))
        output.append(u"%15.15s : %s" %
                      (u"Fichiers à traiter", len(self.list_files())))
        if buff:
            buff.write(u"\n".join(output).encode(ENCODING))
        if log:
            for line in output:
                self.logger.debug(line.encode(ENCODING))
        return output

    def run(self, filecount=None):
        """
        Boucle de processing des objets
        """
        destinations = {
            MOVE_TO_ERROR: os.path.join(self.config['dirfich'], 'erreur'),
            MOVE_TO_DONE: os.path.join(self.config['dirfich'], 'traite'),
        }
        files = self.list_files()
        if not files:
            self.logger.debug(u"Aucun fichier à traiter")
            return
        allfiles = len(files)
        if filecount is not None:
            files = files[:filecount]
        self.logger.debug(
            u"Traitement de %s fichiers sur %s", len(files), allfiles)

        for filename in files:
            full_file_path = os.path.join(self.config['dirfich'], filename)
            self.logger.info(u"%s - Traitement du fichier", filename)
            result = self.process(filename)
            if result == DONT_MOVE:
                self.logger.debug(
                    u"%s - Fin du traitement, le fichier reste en place.",
                    filename)
            elif result == MOVE_TO_TRASH:
                self.logger.debug(
                    u"%s - Fin du traitement, suppression du fichier.",
                    filename)
                os.unlink(full_file_path)
            else:
                wheretomove = destinations[result]
                self.logger.debug(
                    u"%s - Déplacement dans le dossier %s",
                    filename, wheretomove)
                os.rename(full_file_path, os.path.join(wheretomove, filename))

    def process(self, filename):
        """
        Traitement d'un fichier
        """
        self.logger.error(
            u"Rien ne sera fait, il faut sous-classer la classe Worker et "
            u"redéfinir sa méthode 'process'.")
        return DONT_MOVE

    def save_img(self, filename, demande, dem_auto, testbio, bufimg,
                 imgformat='png', automate=None, id_resultat=None,
                 maj_dem_auto_test=False, maj_resultat=False,
                 maj_prescription=False, id_prescription=None):
        """
        Sauve l'image en BDD
        """
        if self.config['compress_img'] or self.config['resize_img']:
            buf = io.BytesIO(bufimg)
            im = Image.open(buf)
            format = im.format
            out = io.BytesIO()
            kwargs = {}
            try:
                # Compression en JPEG
                if self.config["compress_img"]:
                    if format != "JPEG":
                        if im.mode not in ("L", "RGB"):
                            im = im.convert("RGB")
                    format = "JPEG"
                    kwargs['quality'] = 90
                    imgformat = "jpeg"
                # Redimensionnement
                if self.config['resize_img']:
                    self.logger.debug(
                        u"%s - Redimensionnement de l'image en %s",
                        filename, self.config['resize_img'])
                    im.thumbnail(self.config['resize_img'], Image.ANTIALIAS)
                im.save(out, format, **kwargs)
                bufimg = out.read()
                out.close()
            except:
                self.logger.exception("unknown error")
            buf.close()

        if automate is None:
            automate = self.config['paillasse']
        try:
            shadoc_id = None
            if self.config["shadoc_url"]:
                assert id_resultat is not None, (
                    "Shadoc ne marche qu'avec les workers qui fournissent un ID resultat"
                )
                cursor = self.conn.cursor()
                cursor.execute("""
                SELECT
                    d.no_demandeur, d.date_demande,
                    t.code
                FROM demande d, testbio t
                WHERE d.id = %s AND t.id = %s""", (demande, testbio))
                no_demandeur, date_demande, testbio_code = cursor.fetchone()
                self.conn.commit()
                shadoc = shadoc_client.Client(self.config["shadoc_url"])
                shadoc_id = shadoc.upload(
                    no_demande=no_demandeur,
                    date_demande=date_demande,
                    document_type="PJ %s" % (testbio_code,),
                    file=bufimg,
                )["id"]
            buffer_postgres = psycopg2.Binary(bufimg)
            cursor = self.conn.cursor()
            cursor.execute(
                "select max(seq) as max from graphique where dem_auto = %s " +
                "and testbio = %s",
                [dem_auto, testbio])
            themax = cursor.fetchone()
            self.conn.commit()
            if themax and themax[0] is not None:
                cursor.execute(
                    "update graphique set seq = %s where seq=0 and " +
                    "dem_auto = %s and testbio = %s",
                    [themax[0] + 1, dem_auto, testbio])
                self.conn.commit()

            if id_resultat is not None:
                if shadoc_id is not None:
                    cursor.execute("""
                        INSERT INTO graphique (
                            demande, dem_auto, testbio,
                            automate, type_fichier, seq,
                            graphe, resultat_id, shadoc_id
                        ) VALUES (
                            %(demande)s, %(dem_auto)s, %(testbio)s,
                            %(automate)s, %(type_fichier)s, %(seq)s,
                            %(graphe)s, %(resultat_id)s, %(shadoc_id)s
                        ) RETURNING id
                    """, {
                        "demande": demande,
                        "dem_auto": dem_auto,
                        "testbio": testbio,
                        "automate": automate,
                        "type_fichier": imgformat,
                        "seq": 0,
                        "graphe": None,
                        "resultat_id": id_resultat,
                        "shadoc_id": shadoc_id,
                    })
                else:
                    cursor.execute(
                        """insert into graphique
                        (demande, dem_auto ,testbio, automate,
                         type_fichier, seq, graphe, resultat_id)
                        values (%s,%s,%s,%s,%s,%s,%s,%s)
                        returning id
                        """,
                        [demande, dem_auto, testbio, automate,
                         imgformat, 0, buffer_postgres, id_resultat]
                    )
            else:
                cursor.execute(
                    """insert into graphique
                    (demande, dem_auto ,testbio, automate,
                     type_fichier, seq, graphe)
                    values (%s,%s,%s,%s,%s,%s,%s)
                    returning id
                    """,
                    [demande, dem_auto, testbio, automate,
                     imgformat, 0, buffer_postgres]
                )
            id_graphique = cursor.fetchone()
            if maj_dem_auto_test:
                cursor.execute("""
                    update dem_auto_test set status = 4
                    where status in (0,1,2,3,5) and dem_auto = %s
                    and testbio = %s""",
                               [dem_auto, testbio])
            if maj_resultat:
                cursor.execute("""
                    update resultat set stat_resu = 'F', date_resu = now(),
                    resu = %s where demande = %s and testbio = %s
                    and seq = 0""",
                               [filename, demande, testbio])
            if maj_prescription and id_prescription is not None:
                cursor.execute("""
                    update prescription set statut = 2
                    where id = %s""", [id_prescription, ])
            self.conn.commit()
        except:
            self.logger.exception(
                u"%s - erreur à l'insertion du graphique", filename)
            return MOVE_TO_ERROR
        else:
            self.logger.info(u"%s - graphique inséré en BDD sous l'id %s",
                             filename, id_graphique[0])
            if self.config["del_on_success"]:
                self.logger.info(u"%s - del_on_success activé. Suppression du fichier",
                            filename)
                return MOVE_TO_TRASH
            return MOVE_TO_DONE

    def ftpget(self, filename, oid=None, distant_folder=''):
        """
        Récupère un fichier sur le FTP
        """

        # On cherche le fichier dans search_subfolders et on retourne le
        # premier trouvé
        if not distant_folder and self.config["search_subfolders"]:
            for folder in self.config["search_subfolders"].split(","):
                local_filename = self.ftpget(
                    filename, oid=oid, distant_folder=folder.strip())
                if local_filename is not None:
                    return local_filename
            else:
                return None

        filename = os.path.join(distant_folder, filename)

        if not oid:
            oid = filename

        local_ok_fn = None
        if self.config['create_ok'] != '0':
            local_ok_fn = filename.split('.')
            local_ok_fn[-1] = self.config['create_ok']
            local_ok_fn = '.'.join(local_ok_fn)

        self.logger.debug(
            u"%s - Récupération du fichier %s sur le FTP", oid, filename)
        try_again = 5  # Nombre d'essais suite à un timeout
        while try_again > 0:
            try_again -= 1
            ok_fn = None
            if self.config['ftp_ok'] != '0':
                ok_fn = filename.split('.')
                ok_fn[-1] = self.config['ftp_ok']
                ok_fn = '.'.join(ok_fn)
                try:
                    okbuffer = io.BytesIO()
                    self.config['ftp_object'].retrbinary(
                        'RETR ' + ok_fn, okbuffer.write)
                except error_perm:
                    self.logger.warning(
                        u"%s - Pas de fichier OK (%s)", oid, ok_fn)
                    try:
                        self.config['ftp_object'].retrbinary(
                        'RETR ' + os.path.join(distant_folder, self.config['ftp_ok']), okbuffer.write)
                    except:
                        self.logger.warning(
                        u"%s - Pas de fichier semaphore (%s)", oid, os.path.join(distant_folder, self.config['ftp_ok']))
                    return None
                except socket.timeout:
                    self.logger.warning(
                        u"%s - FTP timeout, nouvel essai (%s)", oid, try_again)
                    continue
                except all_errors:
                    self.logger.exception(u"%s - Erreur inconnue", oid)
                    return None
                else:
                    okbuffer = None
            ftp = self.config['ftp_object']
            local_filename = os.path.join(self.config['dirfich'], filename)
            local_folder = os.path.dirname(local_filename)
            tmpfile = None
            try:
                if not os.path.exists(local_folder):
                    # Dans le cas où le fichier distant est dans un
                    # sous répertoire
                    os.makedirs(local_folder)
                tmpfile = tempfile.NamedTemporaryFile(delete=False)
                ftp.retrbinary('RETR ' + filename, tmpfile.write)
                tmpfile.close()
                shutil.move(tmpfile.name, local_filename)
            except error_perm:
                self.logger.error(
                    u"%s - Permission refusée ou fichier %s absent",
                    oid, filename)
                return None
            except socket.timeout:
                self.logger.warning(
                    u"%s - FTP timeout, nouvel essai (%s)", oid, try_again)
                continue
            except all_errors:
                self.logger.exception(u"%s - Erreur inconnue", oid)
                return None
            else:
                if self.config['ftp_del']:
                    self.logger.debug(
                        u"%s - Suppression du fichier %s sur le FTP",
                        oid, filename)
                    try:
                        ftp.delete(filename)
                    except error_perm:
                        self.logger.error(
                            u"%s - Permission refusée sur %s", oid, filename)
                        return None
                    except socket.timeout:
                        self.logger.warning(
                            u"%s - FTP timeout, nouvel essai (%s)",
                            oid, try_again)
                        continue
                    except all_errors:
                        self.logger.exception(u"%s - Erreur inconnue", oid)
                        return None
                    if ok_fn:
                        self.logger.debug(
                            u"%s - Suppression du fichier OK %s sur le FTP",
                            oid, ok_fn)
                        try:
                            ftp.delete(ok_fn)
                        except error_perm:
                            self.logger.error(
                                u"%s - %s - Permission refusée", oid, ok_fn)
                            return None
                        except socket.timeout:
                            self.logger.warning(
                                u"%s - FTP timeout, nouvel essai (%s)",
                                oid, try_again)
                            continue
                        except all_errors:
                            self.logger.exception(
                                u"%s - %s - Erreur inconnue", oid, ok_fn)
                            return None
                if local_ok_fn is not None:
                    local_filename = os.path.join(
                        self.config['dirfich'], local_ok_fn
                    )
                    filepoint = open(local_filename, 'w')
                    filepoint.write("%s\n" % (filename, ))
                    filepoint.close()
                return local_filename
            finally:
                if tmpfile is not None and os.path.exists(tmpfile.name):
                    self.logger.debug(u"Un fichier temporaire a ete cree")
                    os.unlink(tmpfile.name)
        # Si on arrive là c'est qu'on a dépassé le nombre d'essais autorisés
        self.logger.warning(u"%s - Trop d'essais successifs", oid)
        return None

    def ftpput(self, filename, oid=None, distant_folder=''):
        """
        Dépose un fichier sur le FTP
        """
        if not oid:
            oid = filename
        os.path.join(distant_folder, filename)
        local_filename = os.path.join(self.config['dirfich'], filename)
        os.path.dirname(local_filename)
        self.logger.debug(u"%s - Dépôt du fichier sur le FTP", oid)
        try_again = 5  # Nombre d'essais suite à un timeout
        ftp = self.config['ftp_object']
        etape = 0
        while try_again > 0:
            try_again -= 1
            if etape == 0:
                self.logger.debug(
                    u"%s - Envoi du fichier sur le FTP (essais : %s)",
                    oid, try_again)
                filepoint = open(local_filename, 'rb')
                try:
                    ftp.storbinary('STOR ' + filename, filepoint)
                except all_errors:
                    self.logger.exception(
                        u"%s - Erreur de transfert FTP", oid)
                    return MOVE_TO_ERROR
                except socket.timeout:
                    self.logger.warning(
                        u"%s - -FTP timeout, nouvel essai (%s)",
                        oid, try_again)
                    continue
                else:
                    etape = 1
                finally:
                    filepoint.close()
            if etape == 1:
                if self.config['ftp_ok'] != '0':
                    self.logger.debug(
                        u"%s - Envoi d'un fichier OK sur le FTP", oid)
                    ok_fn = filename.split('.')
                    if len(ok_fn) > 1:
                        ok_fn[-1] = self.config['ftp_ok']
                    else:
                        ok_fn.append(self.config['ftp_ok'])
                    ok_fn = '.'.join(ok_fn)
                    okbuffer = io.BytesIO()
                    okbuffer.write(filename.encode("utf-8"))
                    okbuffer.flush()
                    try:
                        ftp.storbinary('STOR ' + ok_fn, okbuffer)
                    except all_errors:
                        self.logger.exception(u"%s - Erreur de transfert FTP", oid)
                        return MOVE_TO_ERROR
                    except socket.timeout:
                        self.logger.warning(
                            u"%s - -FTP timeout, nouvel essai (%s)",
                            oid, try_again)
                        continue
                    finally:
                        okbuffer.close()
                etape = 2
            if etape == 2:
                if self.config['ftp_del']:
                    self.logger.debug(
                        u"%s - Suppression du fichier local", oid)
                    try:
                        os.unlink(local_filename)
                    except:
                        self.logger.exception(u"%s - Impossible de supprimer le fichier")
                        return MOVE_TO_ERROR
                    else:
                        return DONT_MOVE
                etape = 3
            if etape == 3:
                return MOVE_TO_DONE
        self.logger.error(u"%s - Trop de tentatives")
        return MOVE_TO_ERROR

    def list_files(self):
        """
        Liste les fichiers à traiter
        """
        return []

    def list_distant_files(self, folder=None):
        """
        Liste les fichiers distants à traiter
        """
        if not self.config.get('ftp_object', None):
            return
        files = []
        insubfolder = False
        # Bon le protocole FTP ne nous permet pas vraiment de savoir
        # si un 'fichier' est un dossier ou un fichier.
        # Donc on va tester avec CWD :
        #   si on peut rentrer dedans on considère que c'est un dossier
        if folder and not self.config['ftp_recursive']:
            return folder
        if folder:
            try:
                self.config['ftp_object'].cwd(folder)
            except error_perm:
                # Ce n'est pas un dossier
                return folder
            else:
                insubfolder = True
        for filename in self.config['ftp_object'].nlst():
            if filename in ['.', '..']:
                continue
            # Est-ce un dossier ?
            content = self.list_distant_files(filename)
            if content == filename and self.config['ftp_re'].match(filename):
                files.append(filename if folder is None
                             else os.path.join(folder, filename))
            elif isinstance(content, list):
                # La sous-routine nous a renvoyé un résultat
                for subfile in content:
                    files.append(subfile if folder is None
                                 else os.path.join(folder, subfile))

        if insubfolder:
            self.config['ftp_object'].cwd('..')
            return files
        else:
            filtered_files = []
            for filename in files:
                # Si on l'a déjà (on checke juste le nom)
                if not (os.path.exists(os.path.join(
                        self.config['dirfich'], filename))):
                    filtered_files.append(filename)
            return filtered_files

    def list_sftp_distant_files(self, folder=None):
        """
        Liste les fichiers distants à traiter
        """
        sftp = self.config.get('sftp_object', None)

        if not sftp:
            return

        files = []

        if folder:
            content = sftp.listdir_attr(folder)
        else:
            content = sftp.listdir_attr()

        for fichier in content:

            if folder is None:
                path = fichier.filename
            else:
                path = os.path.join(folder, fichier.filename)

            if stat.S_ISDIR(fichier.st_mode):
                files.extend(self.list_sftp_distant_files(path))

            elif self.config['sftp_re'].match(fichier.filename):
                files.append(path)

        return files

    def sftpget(self, filename):
        sftp = self.config['sftp_object']
        local_filename = os.path.join(self.config['dirfich'], filename)
        local_folder = os.path.dirname(local_filename)

        if not os.path.exists(local_folder):
            # Dans le cas où le fichier distant est dans un sous répertoire
            os.makedirs(local_folder)

        # TODO gérer les erreurs...
        sftp.get(filename, local_filename)
        sftp.remove(filename)

        return local_filename

    def list_samba_distant_files(self, dirent=None):
        """
        Liste les fichiers sur le dépôt samba
        """
        ctx = self.config.get('smb_object', None)
        if not ctx:
            return
        files = []
        for direntry in ctx.opendir(self.config['smb_full_url']).getdents():
            filename = direntry.name
            if direntry.smbc_type == 8 and \
                self.config.get('smb_re').match(filename) \
                and not os.path.exists(os.path.join(
                    self.config['dirfich'], filename)) \
                and not any([
                    os.path.exists(os.path.join(
                        self.config['dirfich'], subfolder, filename))
                    for subfolder in self.subfolders]):
                files.append(filename)
        return files

    def smbput(self, filename, oid=None, distant_folder=""):
        """
        Dépose un fichier sur le Samba
        """
        if not oid:
            oid = filename
        local_filename = os.path.join(self.config['dirfich'], filename)
        self.logger.debug(u"%s - Dépôt du fichier sur le SAMBA", oid)
        try_again = 5  # Nombre d'essais suite à un timeout
        open_flags = []
        if self.config['smb_use_flags']:
            open_flags.append(os.O_CREAT | os.O_WRONLY)
        ctx = self.config.get('smb_object', None)
        etape = 0
        while try_again > 0:
            try_again -= 1
            if etape == 0:
                self.logger.debug(
                    u"%s - Envoi du fichier sur le SMB (essais : %s)",
                    oid, try_again)
                filepoint = open(local_filename, 'rb')
                try:
                    remote_file = ctx.open(
                        "%s/%s" % (self.config['smb_full_url'], filename),
                        *open_flags)
                    remote_file.write(filepoint.read())
                except socket.timeout:
                    self.logger.warning(
                        u"%s - -SMB timeout, nouvel essai (%s)",
                        oid, try_again)
                    continue
                except:
                    self.logger.exception(
                        u"%s - Erreur de transfert SAMBA", oid)
                    return MOVE_TO_ERROR
                else:
                    etape = 1
                finally:
                    filepoint.close()
            if etape == 1:
                if self.config['smb_ok'] != '0':
                    self.logger.debug(
                        u"%s - Envoi d'un fichier OK sur le SMB", oid)
                    ok_fn = filename.split('.')
                    if len(ok_fn) > 1:
                        ok_fn[-1] = self.config['smb_ok']
                    else:
                        ok_fn.append(self.config['smb_ok'])
                    ok_fn = '.'.join(ok_fn)
                    okbuffer = io.BytesIO()
                    okbuffer.write(filename.encode("utf-8"))
                    okbuffer.flush()
                    try:
                        remote_file = ctx.open(
                            "%s/%s" % (self.config["smb_full_url"], ok_fn),
                            *open_flags)
                        remote_file.write("")
                    except socket.timeout:
                        self.logger.warning(
                            u"%s - -SMB timeout, nouvel essai (%s)",
                            oid, try_again)
                        continue
                    except:
                        self.logger.exception(u"%s - Erreur de transfert SMB", oid)
                        return MOVE_TO_ERROR
                    finally:
                        okbuffer.close()
                etape = 2
            if etape == 2:
                if self.config['smb_del']:
                    self.logger.debug(
                        u"%s - Suppression du fichier local", oid)
                    try:
                        os.unlink(local_filename)
                    except:
                        self.logger.exception(
                            u"%s - Impossible de supprimer le fichier", oid)
                        return MOVE_TO_ERROR
                    else:
                        return DONT_MOVE
                etape = 3
            if etape == 3:
                return MOVE_TO_DONE
        self.logger.error(u"%s - Trop de tentatives")
        return MOVE_TO_ERROR

    def smbget(self, filename, oid=None):
        """
        Récupère un fichier sur le Samba
        """
        if not oid:
            oid = filename
        ctx = self.config.get('smb_object')
        uri = self.config.get('smb_full_url')
        full_distant_uri = "%s/%s" % (uri, filename)
        try:
            distant_filepoint = ctx.open(full_distant_uri)
        except RuntimeError as e:
            self.logger.warning(
                u"%s - Aucun fichier distant trouvé : %s", oid, e.args)
            return None
        st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime = (
            distant_filepoint.fstat()
        )
        if st_size == 0:
            self.logger.debug(u"%s - fichier Samba vide", filename)
            return None
        if time.time() - st_mtime < 0.5:
            self.logger.debug(u"%s - fichier Samba trop jeune", filename)
            return None

        local_ok_fn = None
        if self.config['create_ok'] != '0':
            local_ok_fn = filename.split('.')
            local_ok_fn[-1] = self.config['create_ok']
            local_ok_fn = '.'.join(local_ok_fn)

        ok_fn = None
        if self.config['smb_ok'] != '0':
            ok_fn = filename.split('.')
            ok_fn[-1] = self.config['smb_ok']
            ok_fn = '.'.join(ok_fn)
            try:
                ctx.open("%s/%s" % (uri, ok_fn))
            except RuntimeError as e:
                self.logger.warning(
                    u"%s - Pas de fichier OK (%s) : %s", oid, ok_fn, e.args)
                return None
        local_filename = os.path.join(self.config['dirfich'], filename)
        local_folder = os.path.dirname(local_filename)
        try:
            if not os.path.exists(local_folder):
                # Dans le cas où le dossier local n'existe pas
                os.makedirs(local_folder)
            filepoint = open(local_filename, 'wb')
            filepoint.write(distant_filepoint.read())
            filepoint.flush()
            os.fsync(filepoint.fileno())
            distant_filepoint.close()
            filepoint.close()
        except RuntimeError as e:
            self.logger.error(
                u"%s - Erreur récupération SAMBA : %s", oid, e.args)
            return None
        if self.config['smb_del']:
            self.logger.debug(
                u"%s - Suppression du fichier %s sur le smb", oid, filename)
            try:
                ctx.unlink(full_distant_uri)
            except RuntimeError as e:
                self.logger.error(
                    u"%s - Erreur suppression SAMBA : %s", oid, e.args)
                return None
            if ok_fn:
                self.logger.debug(
                    u"%s - Suppression du fichier OK %s sur le SAMBA",
                    oid, ok_fn)
                try:
                    ctx.unlink("%s/%s" % (uri, ok_fn))
                except RuntimeError as e:
                    self.logger.error(
                        u"%s - Erreur suppression fichier OK sur SAMBA : %s",
                        oid, e.args)
                    return None
        if local_ok_fn is not None:
            local_filename = os.path.join(
                self.config['dirfich'], local_ok_fn
            )
            filepoint = open(local_filename, 'w')
            filepoint.write("%s\n" % (filename, ))
            filepoint.close()
        return local_filename

    def split_image(self, filename, return_filenames_on_success=False):
        # Lecture du fichier
        if self.config['fname_re']:
            match = self.config['fname_re'].match(filename)
            if not match:
                self.logger.error(
                    u"%s - Le nom du fichier ne matche pas l'expression "
                    u"régulière." % (filename,))
                return MOVE_TO_ERROR
        fichier = os.path.join(self.config['dirfich'], filename)
        if os.path.getsize(fichier) == 0:
            self.logger.error(u"%s - La taille du fichier est zero" % (filename))
            return MOVE_TO_ERROR
        # Découpage
        try:
            im = Image.open(fichier)
        except IOError as e:
            self.logger.error(e)
            return MOVE_TO_ERROR
        self.logger.debug(u"%s - Image de type %s", filename, im.format)
        cropped_filenames = []
        for idx, (left, upper, width, height) in enumerate(
                self.config['split_geometry']):
            right = left + width
            lower = upper + height
            self.logger.debug(
                u"%s - Découper selon left:%d, upper:%d, right:%d, lower:%d",
                filename, left, upper, right, lower)
            cropped = im.crop((left, upper, right, lower))
            cropped.load()
            output_fmt = im.format
            kwargs = {}
            if self.config['compress_img']:
                self.logger.debug(
                    u"%s - Compression de l'image en JPEG", filename)
                output_fmt = "JPEG"
                kwargs['quality'] = 60
            if self.config['resize_img']:
                self.logger.debug(u"%s - Redimensionnement de l'image en %s",
                                  filename, self.config['resize_img'])
                cropped.thumbnail(self.config['resize_img'], Image.ANTIALIAS)

            cropped_filename = os.path.join(
                self.config['outputdir'],
                "%d_%s.%s" % (idx, filename, output_fmt))
            cropped.save(cropped_filename, output_fmt, **kwargs)
            cropped_filenames.append(cropped_filename)
        if return_filenames_on_success:
            return cropped_filenames

        if self.config['del_on_success']:
            self.logger.info(u"%s - del_on_success activé. Suppression du fichier",
                            filename)
            return MOVE_TO_TRASH
        return MOVE_TO_DONE

    def teardown(self):
        if self.config.get("ftp_object") is not None:
            # Certains serveurs FTP custom n'aiment pas qu'on quitte sans
            # envoyer un QUIT...
            self.config["ftp_object"].quit()

        if self.need_bdd and self.conn:
            self.conn.close()
